library slu_defs;

import 'dart:js';
import 'dart:typed_data';

enum ColPerm { NATURAL, MMD_ATA, MMD_AT_PLUS_A, COLAMD }

enum Trans { NOTRANS, TRANS, CONJ }

const _colPermMap = const <ColPerm, int>{
  ColPerm.NATURAL: 0,
  ColPerm.MMD_ATA: 1,
  ColPerm.MMD_AT_PLUS_A: 2,
  ColPerm.COLAMD: 3
};

const _transMap = const <Trans, int>{
  Trans.NOTRANS: 0,
  Trans.TRANS: 1,
  Trans.CONJ: 2
};

class Factors {
  final JsObject _module;
  final String _fname;
  num _pp_factors;

  Factors(JsObject ctx, String moduleName, bool complex, Float64List values, Int32List rowind,
      Int32List colptr, ColPerm permc_spec)
      : _module = (ctx == null ? context : ctx)[moduleName],
        _fname = complex ? '_zlinsol' : '_dlinsol' {
    if (_module == null) {
      throw new ArgumentError.notNull("moduleName");
    }

    int iopt = 1;
    int n = colptr.length - 1;
    int nnz = values.length;
    int nrhs = 0;

    var p_values = _malloc(values);
    var p_rowind = _malloc(rowind);
    var p_colptr = _malloc(colptr);

    var p_b = 0; // NULL
    var ldb = 0;
    var trans = 0;

    int i32 = Int32List.BYTES_PER_ELEMENT; // pointers are 32-bit
    _pp_factors = _module.callMethod('_malloc', [i32]);

    var p_info = _module.callMethod('_malloc', [i32]);
    _module.callMethod('setValue', [p_info, 0, 'i32']);

    _module.callMethod(_fname, [
      iopt,
      n,
      nnz,
      nrhs,
      p_values,
      p_rowind,
      p_colptr,
      p_b,
      ldb,
      _pp_factors,
      _colPermMap[permc_spec],
      trans,
      p_info
    ]);

    freeAll([p_values, p_rowind, p_colptr]);

    var info = _module.callMethod('getValue', [p_info, 'i32']);
    _free(p_info);
    if (info != 0) {
      throw new Exception(info);
    }
  }

  void solve(int n, int nrhs, Float64List rhs,
      {int ldb, Trans trans: Trans.NOTRANS}) {
    const iopt = 2;
    const nnz = 0;
    const p_values = 0;
    const p_rowind = 0;
    const p_colptr = 0;

    var p_b = _malloc(rhs);
    if (ldb == null) {
      ldb = n;
    }

    const permc_spec = 0;

    int i32 = Int32List.BYTES_PER_ELEMENT;
    var p_info = _module.callMethod('_malloc', [i32]);
    _module.callMethod('setValue', [p_info, 0, 'i32']);

    _module.callMethod(_fname, [
      iopt,
      n,
      nnz,
      nrhs,
      p_values,
      p_rowind,
      p_colptr,
      p_b,
      ldb,
      _pp_factors,
      permc_spec,
      _transMap[trans],
      p_info
    ]);

    int numBytes = rhs.length * Float64List.BYTES_PER_ELEMENT;
    Uint8List bytes = _module.callMethod('getTypedData', [p_b, numBytes]);
    rhs.setAll(0, new Float64List.view(bytes.buffer));
    _free(p_b);

    var info = _module.callMethod('getValue', [p_info, 'i32']);
    _free(p_info);
    if (info != 0) {
      throw new Exception(info);
    }
  }

  void free() {
    const iopt = 3;
    const n = 0;
    const nnz = 0;
    const nrhs = 0;
    const p_values = 0;
    const p_rowind = 0;
    const p_colptr = 0;
    const p_b = 0;
    const ldb = 0;
    var pp_factors = _pp_factors;
    const permc_spec = 0;
    const trans = 0;

    int i32 = Int32List.BYTES_PER_ELEMENT;
    var p_info = _module.callMethod('_malloc', [i32]);
    _module.callMethod('setValue', [p_info, 0, 'i32']);

    _module.callMethod(_fname, [
      iopt,
      n,
      nnz,
      nrhs,
      p_values,
      p_rowind,
      p_colptr,
      p_b,
      ldb,
      pp_factors,
      permc_spec,
      trans,
      p_info
    ]);

    var info = _module.callMethod('getValue', [p_info, 'i32']);
    _free(p_info);
    if (info != 0) {
      throw new Exception(info);
    }
  }

  num _malloc(TypedData data) {
    var ptr = _module.callMethod('_malloc', [data.lengthInBytes]);
    _module.callMethod('setTypedData', [data, ptr]);
    return ptr;
  }

  _free(num ptr) => _module.callMethod('_free', [ptr]);

  freeAll(List<num> ptrs) {
    ptrs.forEach((num ptr) => _free(ptr));
  }
}
